@@ -127,6 +127,25 @@ progress_set_format (GladeXML *xml, GtkWidget *w,
     gtk_progress_set_format_string (GTK_PROGRESS (w), value);
 }
 
+static void
+list_item_set_label (GladeXML *xml, GtkWidget *w,
+		     const char *name, const char *value)
+{
+    GtkWidget *child;
+
+    child = gtk_bin_get_child (GTK_BIN (w));
+    if (child)
+	gtk_container_remove (GTK_CONTAINER (w), child);
+    
+    /* copied from gtk_list_item_new_with_label() */
+    child = gtk_label_new (value);
+    gtk_misc_set_alignment (GTK_MISC (child), 0.0, 0.5);
+    gtk_misc_set_padding (GTK_MISC (child), 0, 1);
+
+    gtk_container_add (GTK_CONTAINER (w), child);
+    gtk_widget_show (child);
+}
+
 static void
 text_view_set_text (GladeXML *xml, GtkWidget *w,
 		    const char *name, const char *value)
@@ -362,13 +381,41 @@ menu_item_set_use_stock (GladeXML *xml, GtkWidget *w,
     }
 }
 
-
-static GtkWidget *
-placeholder_create (void)
+static void
+window_set_wmclass_name (GladeXML *xml, GtkWidget *w,
+			 const gchar *name, const gchar *value)
 {
-    GtkWidget *pl = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
-    gtk_widget_show (pl);
-    return pl;
+    GtkWindow *window = GTK_WINDOW(w);
+    gchar *wmclass = g_strdup(window->wmclass_class);
+
+    gtk_window_set_wmclass(window, value, wmclass);
+    g_free(wmclass);
+}
+
+static void
+window_set_wmclass_class (GladeXML *xml, GtkWidget *w,
+			  const gchar *name, const gchar *value)
+{
+    GtkWindow *window = GTK_WINDOW(w);
+    gchar *wmname = g_strdup(window->wmclass_name);
+
+    gtk_window_set_wmclass(window, wmname, value);
+    g_free(wmname);
+}
+
+static void
+entry_set_invisible_char (GladeXML *xml, GtkWidget *w,
+			  const gchar *name, const gchar *value)
+{
+    gtk_entry_set_invisible_char (GTK_ENTRY (w), value [0]);
+}
+
+static void
+button_set_response_id (GladeXML *xml, GtkWidget *w,
+			const gchar *name, const gchar *value)
+{
+    g_object_set_data (G_OBJECT (w), "response_id",
+		       GINT_TO_POINTER (INT (value)));
 }
 
 static void
@@ -410,22 +457,64 @@ gtk_dialog_build_children(GladeXML *self, GtkWidget *w,
     if (dialog->action_area == NULL)
 	return;
 
-    children = gtk_container_get_children (GTK_CONTAINER (dialog->action_area));
+    /* repack children of action_area */
+    children = gtk_container_get_children(GTK_CONTAINER(dialog->action_area));
     for (list = children; list; list = list->next) {
-	gtk_widget_ref (GTK_WIDGET (list->data));
-	gtk_container_remove (GTK_CONTAINER (dialog->action_area), GTK_WIDGET (list->data));
+	GtkWidget *child = GTK_WIDGET(list->data);
+
+	g_object_ref(child);
+	gtk_container_remove (GTK_CONTAINER (dialog->action_area), child);
     }
+    for (list = children; list; list = list->next) {
+	GtkWidget *child = GTK_WIDGET(list->data);
+	gint response_id;
+
+	response_id = GPOINTER_TO_INT(g_object_steal_data(G_OBJECT(child),
+							  "response_id"));
+	gtk_dialog_add_action_widget(dialog, child, response_id);
+	g_object_unref(child);
 
-    for (list = children; list; list = list->next) {
-	gint response_id;
-	response_id = GPOINTER_TO_INT (g_object_steal_data (
-	    G_OBJECT (list->data), "response_id"));
-	gtk_dialog_add_action_widget (dialog, GTK_WIDGET (list->data), response_id);
     }
-    g_list_foreach (children, (GFunc)gtk_widget_unref, NULL);
     g_list_free (children);
 }
 
+
+static void
+frame_build_children(GladeXML *self, GtkWidget *parent,
+		     GladeWidgetInfo *info)
+{
+    gint i, j;
+    enum {
+	FRAME_ITEM,
+	LABEL_ITEM
+    } type;
+
+    g_object_ref(G_OBJECT(parent));
+    for (i = 0; i < info->n_children; i++) {
+	GladeWidgetInfo *childinfo = info->children[i].child;
+	GtkWidget *child = glade_xml_build_widget(self, childinfo);
+
+	type = FRAME_ITEM;
+	/* there should really only be 2 children */
+	for (j = 0; j < info->children[i].n_properties; j++) {
+	    if (!strcmp (info->children[i].properties[j].name, "type")) {
+		const char *value = info->children[i].properties[j].value;
+
+		if (!strcmp (value, "label_item"))
+		    type = LABEL_ITEM;
+		break;
+	    }
+	}
+
+	if (type == LABEL_ITEM) {
+	    gtk_frame_set_label_widget (GTK_FRAME (parent), child);
+	} else {
+	    gtk_container_add (GTK_CONTAINER (parent), child);
+	}
+    }
+    g_object_unref(G_OBJECT(parent));
+}
+
 static void
 notebook_build_children(GladeXML *self, GtkWidget *parent,
 			GladeWidgetInfo *info)
@@ -499,7 +588,19 @@ option_menu_build_children (GladeXML *xml, GtkWidget *parent,
 {
     int i, history = 0;
 
-    glade_standard_build_children (xml, parent, info);
+    for (i = 0; i < info->n_children; i++) {
+	GtkWidget *child;
+	GladeWidgetInfo *childinfo = info->children[i].child;
+
+	if (strcmp(childinfo->classname, "GtkMenu") != 0) {
+	    g_warning("the child of the option menu '%s' was not a GtkMenu",
+		      info->name);
+	    continue;
+	}
+	child = glade_xml_build_widget(xml, childinfo);
+
+	gtk_option_menu_set_menu(GTK_OPTION_MENU(parent), child);
+    }
 
     for (i = 0; i < info->n_properties; i++) {
 	const char *name  = info->properties[i].name;
@@ -530,7 +631,7 @@ clist_build_children(GladeXML *self, GtkWidget *parent,
 	childinfo = info->children[i].child;
 
 	/* treat GtkLabels specially */
-	if (!strcmp (childinfo->class, "GtkLabel")) {
+	if (!strcmp (childinfo->classname, "GtkLabel")) {
 	    int j;
 	    const char *label = NULL;
 
@@ -575,9 +676,9 @@ toolbar_build_children (GladeXML *xml, GtkWidget *parent,
 
 	childinfo = &info->children[i];
 
-	if (!strcmp (childinfo->child->class, "toggle") ||
-	    !strcmp (childinfo->child->class, "radio") ||
-	    !strcmp (childinfo->child->class, "button")) {
+	if (!strcmp (childinfo->child->classname, "toggle") ||
+	    !strcmp (childinfo->child->classname, "radio") ||
+	    !strcmp (childinfo->child->classname, "button")) {
 	    const char *label = NULL, *stock = NULL, *group_name = NULL;
 	    char *icon = NULL;
 	    gboolean use_stock = FALSE, active = FALSE, new_group = FALSE;
@@ -643,6 +744,7 @@ toolbar_build_children (GladeXML *xml, GtkWidget *parent,
 		    g_warning ("Could not create stock item: %s", stock);
 	    } else if (icon) {
 		iconw = gtk_image_new_from_file (icon);
+		g_free (icon);
 	    }
 
 	    if (iconw)
@@ -652,14 +754,14 @@ toolbar_build_children (GladeXML *xml, GtkWidget *parent,
 		gtk_toolbar_append_space (GTK_TOOLBAR (parent));
 
 	    /* FIXME: these should be translated */
-	    if (!strcmp (childinfo->child->class, "toggle")) {
+	    if (!strcmp (childinfo->child->classname, "toggle")) {
 		child = gtk_toolbar_append_element (
 		    GTK_TOOLBAR (parent),
 		    GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL,
 		    label, NULL, NULL, iconw, NULL, NULL);
 		gtk_toggle_button_set_active(
 		    GTK_TOGGLE_BUTTON (child), active);
-	    } else if (!strcmp (childinfo->child->class, "radio")) {
+	    } else if (!strcmp (childinfo->child->classname, "radio")) {
 		child = gtk_toolbar_append_element (
 		    GTK_TOOLBAR (parent),
 		    GTK_TOOLBAR_CHILD_RADIOBUTTON, NULL,
@@ -743,27 +845,37 @@ paned_build_children (GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
     gtk_paned_pack2 (GTK_PANED(w), child, resize, shrink);
 }
 
-static GtkWidget *
-build_button(GladeXML *xml, GType widget_type,
-	     GladeWidgetInfo *info)
+static void
+layout_build_children (GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
 {
-    GtkWidget *widget;
-    gint i, response_id = 0;
+    gint i;
 
-    for (i = 0; i < info->n_properties; i++) {
-	if (!strcmp (info->properties[i].name, "response_id")) {
-	    response_id = INT (info->properties[i].value);
-	    break;
+    for (i = 0; i < info->n_children; i++) {
+	GladeWidgetInfo *childinfo = info->children[i].child;
+	GtkWidget *child = NULL;
+	gint j, x = 0, y = 0;
+
+	if (info->children[i].internal_child) {
+	    glade_xml_handle_internal_child(xml, w, &info->children[i]);
+	    continue;
 	}
+
+	child = glade_xml_build_widget(xml, childinfo);
+	for (j = 0; j < info->children[i].n_properties; j++) {
+	    const gchar *name = info->children[i].properties[j].name;
+	    const gchar *value = info->children[i].properties[j].value;
+
+	    if (name[0] == 'x' && name[1] == '\0')
+		x = INT(value);
+	    else if (name[0] == 'y' && name[1] == '\0')
+		y = INT(value);
+	    else
+		g_warning("unknown child packing property %s for GtkLayout",
+			  name);
+	}
+
+	gtk_layout_put(GTK_LAYOUT(w), child, x, y);
     }
-
-    widget = glade_standard_build_widget (xml, widget_type, info);
-
-    if (response_id)
-	g_object_set_data (G_OBJECT (widget), "response_id",
-			   GINT_TO_POINTER (response_id));
-
-    return widget;
 }
 
 static GtkWidget *
@@ -798,26 +910,6 @@ image_menu_find_internal_child(GladeXML *xml, GtkWidget *parent,
     return NULL;
 }
 
-static GtkWidget *
-option_menu_find_internal_child(GladeXML *xml, GtkWidget *parent,
-				const gchar *childname)
-{
-    if (!strcmp(childname, "menu")) {
-	GtkWidget *ret;
-	
-	if ((ret = gtk_option_menu_get_menu (GTK_OPTION_MENU (parent))))
-	    return ret;
-
-	ret = gtk_menu_new ();
-	gtk_widget_show (ret);
-	gtk_option_menu_set_menu (GTK_OPTION_MENU (parent), ret);
-
-	return ret;
-    }
-
-    return NULL;
-}
-
 static GtkWidget *
 scrolled_window_find_internal_child(GladeXML *xml, GtkWidget *parent,
 				    const gchar *childname)
@@ -934,6 +1026,11 @@ _glade_init_gtk_widgets(void)
     glade_register_custom_prop (GTK_TYPE_MENU_ITEM, "label", menu_item_set_label);
     glade_register_custom_prop (GTK_TYPE_MENU_ITEM, "use_underline", menu_item_set_use_underline);
     glade_register_custom_prop (GTK_TYPE_MENU_ITEM, "use_stock", menu_item_set_use_stock);
+    glade_register_custom_prop (GTK_TYPE_WINDOW, "wmclass_name", window_set_wmclass_name);
+    glade_register_custom_prop (GTK_TYPE_WINDOW, "wmclass_class", window_set_wmclass_class);
+    glade_register_custom_prop (GTK_TYPE_LIST_ITEM, "label", list_item_set_label);
+    glade_register_custom_prop (GTK_TYPE_BUTTON, "response_id", button_set_response_id);
+    glade_register_custom_prop (GTK_TYPE_ENTRY, "invisible_char", entry_set_invisible_char);
 
     glade_register_widget (GTK_TYPE_ACCEL_LABEL, glade_standard_build_widget,
 			   NULL, NULL);
@@ -942,8 +1039,8 @@ _glade_init_gtk_widgets(void)
     glade_register_widget (GTK_TYPE_ARROW, glade_standard_build_widget,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_ASPECT_FRAME, glade_standard_build_widget,
-			   glade_standard_build_children, NULL);
-    glade_register_widget (GTK_TYPE_BUTTON, build_button,
+			   frame_build_children, NULL);
+    glade_register_widget (GTK_TYPE_BUTTON, glade_standard_build_widget,
 			   glade_standard_build_children, NULL);
     glade_register_widget (GTK_TYPE_CALENDAR, glade_standard_build_widget,
 			   NULL, NULL);
@@ -980,7 +1077,7 @@ _glade_init_gtk_widgets(void)
     glade_register_widget (GTK_TYPE_FONT_SELECTION_DIALOG, NULL,
 			   glade_standard_build_children, fontseldlg_find_internal_child);
     glade_register_widget (GTK_TYPE_FRAME, glade_standard_build_widget,
-			   glade_standard_build_children, NULL);
+			   frame_build_children, NULL);
     glade_register_widget (GTK_TYPE_GAMMA_CURVE, glade_standard_build_widget,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_HANDLE_BOX, glade_standard_build_widget,
@@ -1008,7 +1105,7 @@ _glade_init_gtk_widgets(void)
     glade_register_widget (GTK_TYPE_LABEL, glade_standard_build_widget,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_LAYOUT, glade_standard_build_widget,
-			   glade_standard_build_children, NULL);
+			   layout_build_children, NULL);
     glade_register_widget (GTK_TYPE_LIST, glade_standard_build_widget,
 			   glade_standard_build_children, NULL);
     glade_register_widget (GTK_TYPE_LIST_ITEM, glade_standard_build_widget,
@@ -1024,11 +1121,13 @@ _glade_init_gtk_widgets(void)
     glade_register_widget (GTK_TYPE_NOTEBOOK, glade_standard_build_widget,
 			   notebook_build_children, NULL);
     glade_register_widget (GTK_TYPE_OPTION_MENU, glade_standard_build_widget,
-			   option_menu_build_children, option_menu_find_internal_child);
+			   option_menu_build_children, NULL);
     glade_register_widget (GTK_TYPE_PIXMAP, glade_standard_build_widget,
 			   NULL, NULL);
+#ifdef HAVE_GTK_PLUG
     glade_register_widget (GTK_TYPE_PLUG, NULL,
 			   NULL, NULL);
+#endif
     glade_register_widget (GTK_TYPE_PREVIEW, build_preview,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_PROGRESS, glade_standard_build_widget,
@@ -1042,8 +1141,12 @@ _glade_init_gtk_widgets(void)
     glade_register_widget (GTK_TYPE_SCROLLED_WINDOW, glade_standard_build_widget,
 			   glade_standard_build_children,
 			   scrolled_window_find_internal_child);
+    glade_register_widget (GTK_TYPE_SEPARATOR_MENU_ITEM, glade_standard_build_widget,
+			   NULL, NULL);
+#ifdef HAVE_GTK_PLUG
     glade_register_widget (GTK_TYPE_SOCKET, glade_standard_build_widget,
 			   NULL, NULL);
+#endif
     glade_register_widget (GTK_TYPE_SPIN_BUTTON, glade_standard_build_widget,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_STATUSBAR, glade_standard_build_widget,
