@@ -1165,6 +1165,11 @@ glade_parser_parse_file(const gchar *file, const gchar *domain)
 {
     GladeParseState state = { 0 };
 
+    if (!g_file_test(file, G_FILE_TEST_IS_REGULAR)) {
+	g_warning("could not find glade file '%s'", file);
+	return NULL;
+    }
+
     state.interface = NULL;
     if (domain)
 	state.domain = domain;
@@ -1172,13 +1177,13 @@ glade_parser_parse_file(const gchar *file, const gchar *domain)
 	state.domain = textdomain(NULL);
 
     if (xmlSAXUserParseFile(&glade_parser, &state, file) < 0) {
-	g_warning("document not well formed!");
+	g_warning("document not well formed");
 	if (state.interface)
 	    glade_interface_destroy (state.interface);
 	return NULL;
     }
     if (state.state != PARSER_FINISH) {
-	g_warning("did not finish in PARSER_FINISH state!");
+	g_warning("did not finish in PARSER_FINISH state");
 	if (state.interface)
 	    glade_interface_destroy(state.interface);
 	return NULL;
